/* -*- c++ -*- ***************************************************************/
/* Echtzeitsysteme                                                           */
/*---------------------------------------------------------------------------*/
/*                                                                           */
/*                                    T M                                    */
/*                                                                           */
/*---------------------------------------------------------------------------*/
/* The 16-bit timers of the GBA                                              */
/*---------------------------------------------------------------------------*/

#ifndef __devices_tm_h__
#define __devices_tm_h__

#include "devices/derivate.h"
#include "devices/interrupt_controller.h"
#include "devices/interrupt_timer.h"
#include "infra/memory_mapped.h"
#include "infra/types.h"
#include "interrupt/plugbox.h"

/**
 * \class TM
 * \brief The baseclass for the 16-bit timers of the GBA
 **/
class TM : public Interrupt_Timer {

  ezstubs_uint16 *TMCNT_L;
  ezstubs_uint16 *TMCNT_H;

  ezstubs_uint32 VECTOR;
  /**
   * \brief The prescalers available for the GBA timers
   **/
  static const ezstubs_uint32 prescalers[];

  /**
   * \brief Does the timer currently periodically generate interrupts or not
   **/
  bool periodic;

  /**
   * \brief Need to correct time measurement results.
   *
   * The GBA Timers support the periodic generation of interrupts only via a 
   * reload value, that is written into the counter register when the counter
   * overflows. This is problematic when an interrupt occurs during time 
   * measurement, especially when time measurement is startet in a isr/dsr and
   * stopped in a subsequent activation of the isr/dsr.
   * At occurrence of the interrupt the reload value is written into the counter
   * register, so the recorded start and stop values of the measurement will not
   * differ a lot, except the reload value changed in the meanwhile. All in all
   * the measurement is useless. In order to facilitate time measurements also
   * across interrupts, we maintain this variable to record the time that has
   * vanished so far. This variable is updated by acknowledge() (as this method
   * is called on every interrupt) and taken into account by get_time().
   *
   * \see reload
   * \see start_time()
   * \see get_time()
   * \see acknowledge()
   **/
  tick_type correction_ticks;

  /**
   * \brief The current reload value of the counter
   *
   * In order to update the variable correction_ticks appropriately we have
   * to know the current reload value of the timer.
   *
   * \see correction_ticks
   * \see start_time()
   * \see get_time()
   * \see acknowledge()
   **/
  tick_type reload;

protected:

  /**
   * \brief Constructor
   **/
  TM(ezstubs_uint32 offset,ezstubs_uint32 vector) 
    : Interrupt_Timer(vector),
      TMCNT_L((ezstubs_uint16*)offset),
      TMCNT_H((ezstubs_uint16*)(offset + 0x2)),
      VECTOR(vector),
      periodic(false),
      correction_ticks(0) 
  {}

public:

  tick_type max_value();
  tick_type value();

  void start();
  void stop();

  ns_type tick(ns_type ns);

  us_type period();
  us_type period(us_type period);

  ns_type min_period();
  us_type max_period();

  /**
   * \brief Start time measurement
   *
   * Every time we start a time measurement, we can reset correction_ticks.
   * start_time() marks the beginning of time measurement, so no interrupt has
   * been occured during this time measurement so far.
   *
   * \see reload
   * \see correction_ticks
   * \see get_time()
   * \see acknowledge()
   **/
  void start_time();
  us_type get_time();

  /**
   * \brief Trigger the generation of an interrupt request
   *
   * This method is needed to implement the functionality of
   * a test interrupt
   **/
  void trigger();

  /**
   * \brief Check whether an interrupt request has been generated
   *
   * This method is needed to implement the functionality of
   * a test interrupt
   **/
  bool is_triggered();

  void trigger_single();
  void trigger_periodical();
  void cancel();

  /**
   * \brief Reset the interrupt request
   *
   * As the GBA Timers do not support single shot operation in hardware, we
   * emulate this by a flag. If the timer is supposed to generate a single
   * interrupt, we cancel interrupt generation after the first interrupt.
   *
   * For correct time measurement we have to update correction_ticks here.
   * We accumulate the ticks that elapsed since the call to start_time or
   * the last interrupt. We set start_value to reload as counting operation
   * begins from the reload value after an overflow (i.e. after an interrupt)
   *
   * \see reload
   * \see correction_ticks
   * \see start_time()
   * \see get_time()
   * \see acknowledge()
   **/
  void acknowledge();
};

#endif /* __devices_tm_h__ */
